--  2001 Peter Thiemann
module HTMLBase 
  (ATTR_(), attr_, attr_name, attr_value
  ,ELEMENT_(), element_, empty_, cdata_, comment_, doctype_
  ,CDATA_OPTIONS(..)
  ,add_, add_attr_
  ,putElement)
where

import Char

-- untyped layer
-- attributes

data ATTR_ =
  ATTR_ { attr_name :: String
        , attr_value :: String
	}

attr_ = ATTR_

-- elements

data ELEMENT_ =
    ELEMENT_ { tag :: String
    	     , attrs :: [ATTR_]
	     , elems :: [ELEMENT_]
	     }
  | EMPTY_   { tag :: String
  	     , attrs :: [ATTR_]
	     }
  | CDATA_ String
  | COMMENT_ String
  | DOCTYPE_ { doctype :: [String]
  	     , elems :: [ELEMENT_]
	     }

data CDATA_OPTIONS = CDATA_ENCODED | CDATA_FORMATTED
  deriving (Eq)

element_ = ELEMENT_
empty_ = EMPTY_
cdata_ opt = CDATA_ . format . encode
  where format | CDATA_FORMATTED `elem` opt = id
	       | otherwise = htmlFormat
	encode | CDATA_ENCODED `elem` opt = id
	       | otherwise = htmlEncode
comment_ = COMMENT_
doctype_ = DOCTYPE_

add_ e_ e'_ = e_ { elems = e'_ : elems e_}
add_attr_ e_ att = e_ { attrs = att : attrs e_ }

-- show functions

instance Show ATTR_ where
  show = show_attribute
  showList xs s = show_attributes xs ++ s

show_attributes :: [ATTR_] -> String
show_attributes atts = concat (map show_attribute atts)

show_attribute :: ATTR_ -> String
show_attribute a =
  ' ' : attr_name a ++
  case attr_value a of
    "()" -> ""
    str@('\"':_) -> "=" ++ str
    str -> "=\"" ++ str ++ "\""

instance Show ELEMENT_ where
  show = show_element
  showList xs s = show_elements xs ++ s

show_elements :: [ELEMENT_] -> String
show_elements elts = concat (reverse (map show_element elts))

show_element :: ELEMENT_ -> String
show_element (EMPTY_ tag atts) =
  "<" ++ tag ++ show_attributes atts ++ "\n/>"
show_element (ELEMENT_ tag atts elts) =
  "<" ++ tag ++ show_attributes atts ++ ">"
  ++ show_elements elts
  ++ "</" ++ tag ++ "\n>"
show_element (DOCTYPE_ strs elems) =
  "<!DOCTYPE" ++ foldr (++) [] (map (' ' :) strs) ++ "\n>" ++
  "<!-- generated by WASH/HTML 0.9\n-->" ++
  show_elements elems ++ "\n"
show_element (CDATA_ str) =
  str
show_element (COMMENT_ str) =
  "<!-- " ++ str ++ "\n-->"

htmlEncode "" = ""
htmlEncode (x:xs) = 
	case x of
	  '&' -> "&amp;" ++ htmlEncode xs
	  '<' -> "&lt;"  ++ htmlEncode xs
	  '>' -> "&gt;"  ++ htmlEncode xs
	  '\"' -> "&quot;" ++ htmlEncode xs
	  _ -> x : htmlEncode xs

htmlFormat s = s
{--htmlFormat "" = ""
htmlFormat (x:xs) =
	if isSpace x then
	  "\n " ++ htmlFormat (dropWhile isSpace xs)
	else
	  x : htmlFormat xs	  
--}

-- IO functions

putAttribute :: ATTR_ -> IO ()
putAttribute attr =
  putStr (show attr)

openingTag tag attrs =
     do putChar '<'
	putStr tag
	mapM_ putAttribute attrs
	putChar '>'

singleTag tag attrs =
    do putChar '<'
       putStr tag
       mapM_ putAttribute attrs
       putStr " />"

closingTag tag =
     do putStr "</"
	putStr tag
	putStr "\n>"

putElement :: ELEMENT_ -> IO ()
putElement (ELEMENT_ name attrs docs) =
  do openingTag name attrs
     mapM_ putElement (reverse docs)
     closingTag name
putElement (EMPTY_ name attrs) =
  do singleTag name attrs
     putStrLn ""
putElement elem@(CDATA_ text) =
  putStr $ show elem
putElement elem@(COMMENT_ text) =
  putStr $ show elem
putElement (DOCTYPE_ strs elems) =
  do putStr "<!DOCTYPE"
     mapM_ (putStr . (' ' : )) strs
     putStr "\n>"
     putStr "<!-- generated by WASH/HTML 0.9\n-->"
     mapM_ putElement elems
     putStrLn ""

